import logging
from typing import Any, Dict, Optional

from django.utils.translation import gettext_lazy as _

import jwt
from rest_framework import exceptions

from .models import ApplicationCredentials

logger = logging.getLogger(__name__)


ALG = "HS256"


class ZGWAuth:
    def __init__(self, encoded: str):
        self.encoded = encoded

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.payload)

    @property
    def payload(self) -> Optional[Dict[str, Any]]:
        if self.encoded is None:
            return None

        if not hasattr(self, "_payload"):
            # decode the JWT and validate it

            # jwt check
            try:
                payload = jwt.decode(
                    self.encoded,
                    options={"verify_signature": False},
                    algorithms=[ALG],
                )
            except jwt.DecodeError:
                logger.info("Invalid JWT encountered")
                raise exceptions.AuthenticationFailed(
                    _(
                        "JWT could not be decoded. Possibly you made a copy-paste mistake."
                    ),
                    code="jwt-decode-error",
                )

            # get client_id
            try:
                client_id = payload["client_id"]
            except KeyError:
                raise exceptions.AuthenticationFailed(
                    _("`client_id` claim is missing in the JWT."),
                    code="missing-client-identifier",
                )

            # find client_id in DB and retrieve its secret
            try:
                jwt_secret = ApplicationCredentials.objects.exclude(secret="").get(
                    client_id=client_id
                )
            except ApplicationCredentials.DoesNotExist:
                raise exceptions.AuthenticationFailed(
                    _("Client identifier does not exist"),
                    code="invalid-client-identifier",
                )
            else:
                key = jwt_secret.secret

            # check signature of the token
            try:
                payload = jwt.decode(
                    self.encoded,
                    key,
                    algorithms=[ALG],
                )
            except jwt.InvalidSignatureError:
                logger.exception("Invalid signature - possible payload tampering?")
                raise exceptions.AuthenticationFailed(
                    _("Client credentials are invalid."), code="invalid-jwt-signature"
                )

            self._payload = payload

        return self._payload



from drf_spectacular.extensions import OpenApiAuthenticationExtension


class ZGWAuthenticationScheme(OpenApiAuthenticationExtension):
    target_class = "zgw_auth_backend.authentication.ZGWAuthentication"
    name = "ZGWAuthentication"

    def get_security_definition(self, auto_schema):
        return {
            "type": "http",
            "in": "beader",
            "bearerFormat": "JWT",
        }



from django.contrib import admin

from .models import ApplicationCredentials


@admin.register(ApplicationCredentials)
class ApplicationCredentialsAdmin(admin.ModelAdmin):
    list_display = ("client_id",)
    search_fields = ("clien_id",)



# Generated by Django 2.2.13 on 2020-12-15 15:31

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ApplicationCredentials",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "client_id",
                    models.CharField(
                        help_text="Client ID to identify external API's and applications that access this API.",
                        max_length=50,
                        unique=True,
                        verbose_name="client ID",
                    ),
                ),
                (
                    "secret",
                    models.CharField(
                        help_text="Secret belonging to the client ID.",
                        max_length=255,
                        verbose_name="secret",
                    ),
                ),
            ],
            options={
                "verbose_name": "client credential",
                "verbose_name_plural": "client credentials",
            },
        ),
    ]



# Mostly taken from https://github.com/VNG-Realisatie/vng-api-common/blob/master/vng_api_common/middleware.py
#
# We can't use vng-api-common at the moment because of the hard pinned dependencies on
# DRF 3.10 and drf-yasg 1.16.0
import logging

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework.request import Request

from .zgw import ZGWAuth

logger = logging.getLogger(__name__)


class ZGWAuthentication(BaseAuthentication):
    www_authenticate_realm = "api"

    def authenticate(self, request: Request):
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != b"bearer":
            return None

        if len(auth) == 1:
            msg = _("Invalid bearer header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _(
                "Invalid bearer header. Credentials string should not contain spaces."
            )
            raise exceptions.AuthenticationFailed(msg)

        auth = ZGWAuth(auth[1].decode("utf-8"))

        user_id = auth.payload.get("user_id")
        if not user_id:
            msg = _("Invalid 'user_id' claim. The 'user_id' should not be empty.")
            raise exceptions.AuthenticationFailed(msg)

        email = auth.payload.get("email", "")
        return self.authenticate_user_id(user_id, email)

    def authenticate_user_id(self, username: str, email: str):
        UserModel = get_user_model()
        fields = {UserModel.USERNAME_FIELD: username}
        user, created = UserModel._default_manager.get_or_create(**fields)
        if created:
            msg = "Created user object for username %s" % username
            logger.info(msg)

        if email:
            email_field = UserModel.get_email_field_name()
            email_value = getattr(user, email_field)
            if not email_value or email_value != email:
                setattr(user, email_field, email)
                user.save()
                msg = "Set email to %s of user with username %s" % (email, username)
                logger.info(msg)

        return (user, None)

    def authenticate_header(self, request):
        return 'Bearer realm="%s"' % self.www_authenticate_realm



from django.db import models
from django.utils.translation import gettext_lazy as _


class ApplicationCredentials(models.Model):
    client_id = models.CharField(
        _("client ID"),
        max_length=50,
        unique=True,
        help_text=_(
            "Client ID to identify external API's and applications that access this API."
        ),
    )
    secret = models.CharField(
        _("secret"), max_length=255, help_text=_("Secret belonging to the client ID.")
    )

    class Meta:
        verbose_name = _("client credential")
        verbose_name_plural = _("client credentials")

    def __str__(self):
        return self.client_id



from django.apps import AppConfig, apps


class ZgwAuthBackendConfig(AppConfig):
    name = "zgw_auth_backend"

    def ready(self):
        register_spectacular_extensions()


def register_spectacular_extensions():
    if not apps.is_installed("drf_spectacular"):
        return

    from .contrib import drf_spectacular  # noqa
